from datetime import datetime
from typing import Any, Dict, List

from PS3838._utils.errors import ConfigError, ParameterError


def check_config(config : Dict[str, str] = None) -> None:
    if not config:
        raise ConfigError("Config is missing")
    if not isinstance(config, dict):
        raise ConfigError("Config must be a dictionary")
    if not all(key in config for key in ["telegram", "ps3838"]):
        raise ConfigError("Config must contain keys \"telegram\" and \"ps3838\", even if they are empty")



def check_list_matches(list_matches : List[Dict[str, Any]] = None) -> None:
    # check if the list of matches is provided
    if not list_matches:
        raise ParameterError("The list of matches is missing")
    
    # check if the list of matches is a list
    if not isinstance(list_matches, list):
        raise ParameterError("The list of matches must be a list")
    
    # check each match in the list of matches
    for match in list_matches:
        check_match(match)


def check_match(match : Dict[str, Any]) -> None:
    if not isinstance(match, dict):
        raise ParameterError("Each match must be a dictionary")
    if "id" not in match.keys():
        raise ParameterError(f"Parameter \"id\" is missing for at least one match")
    if "starts" not in match.keys():
        raise ParameterError(f"Parameter \"starts\" is missing for at least one match")
    if "league" not in match.keys():
        raise ParameterError(f"Parameter \"league\" is missing for at least one match")
    if "home" not in match.keys():
        raise ParameterError(f"Parameter \"team1\" is missing for at least one match")
    if "away" not in match.keys():
        raise ParameterError(f"Parameter \"team2\" is missing for at least one match")
    if "line_id" not in match.keys():
        raise ParameterError(f"Parameter \"line_id\" is missing for at least one match")
    if "odds" not in match.keys():
        raise ParameterError(f"Parameter \"odds\" is missing for at least one match")
    if "result" not in match.keys():
        raise ParameterError(f"Parameter \"result\" is missing for at least one match")
    if "amount" not in match.keys():
        raise ParameterError(f"Parameter \"amount\" is missing for at least one match")
    if type(match["id"]) != int:
        raise ParameterError(f"\"id\" for match  {match['home']} - {match['away']}  must be an integer")
    if type(match["starts"]) != str:
        raise ParameterError(f"\"starts\" for match  {match['home']} - {match['away']}  must be a datetime object")
    if type(match["league"]) != int:
        raise ParameterError(f"\"league\" for match  {match['home']} - {match['away']}  must be an integer")
    if type(match["home"]) != str:
        raise ParameterError(f"\"home\" for match  {match['home']} - {match['away']}  must be a string")
    if type(match["away"]) != str:
        raise ParameterError(f"\"away\" for match  {match['home']} - {match['away']}  must be a string")
    if type(match["line_id"]) != int:
        raise ParameterError(f"\"line_id\" for match  {match['home']} - {match['away']}  must be an integer")
    if type(match["odds"]) != dict or (type(match["odds"]) == dict and not all(key in match["odds"] for key in ["home", "away", "draw"])):
        raise ParameterError(f"\"odds\" for match  {match['home']} - {match['away']}  must be a dictionary with keys \"home\", \"away\" and \"draw\"")
    if type(match["result"]) != int or (type(match["result"]) == int and match["result"] not in [0, 1, 2]):
        raise ParameterError(f"Result for match  {match['home']} - {match['away']}  must be an integer in [0, 1, 2]")
    if (type(match["amount"]) != float and type(match["amount"]) != int) or (((type(match["amount"]) == int) or (type(match["amount"]) == float)) and match["amount"] < 5):
        raise ParameterError(f"Amount for match  {match['home']} - {match['away']}  must be an integer or a float greater than 5")
        



import logging
from typing import Dict

import requests


class CustomLogger:
    """
    This class creates a custom logger that can log to a file and send messages to a Telegram chat and a .log file.
    """
    def __init__(
            self, 
            name : str, 
            log_file : str = None, 
            func : str = "retrieving", 
            config : Dict[str, str] = None,
            *args, 
            **kwargs
        ):
        self.logger = logging.getLogger(name)
        self.logger.setLevel(logging.DEBUG)
        
        # File handler
        if log_file:
            file_formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
            file_handler = logging.FileHandler(log_file)
            file_handler.setLevel(logging.DEBUG) # Log everything to the file

            # Insert a newline before starting a new logging session (only for clarity)
            with open(log_file, 'a') as f:
                f.write('\n')

            file_handler.setFormatter(file_formatter)
            self.logger.addHandler(file_handler)
        
        # Telegram handler
        if 'telegram_token' in config.keys() and 'telegram_chat_id' in config.keys():
            telegram_formatter = logging.Formatter('%(name)s - %(levelname)s - %(message)s')
            telegram_handler = TelegramBotHandler(config['telegram_token'], config['telegram_chat_id'], func)
            telegram_handler.setLevel(logging.INFO)  # Only INFO and above will be sent to Telegram
            telegram_handler.setFormatter(telegram_formatter)
            self.logger.addHandler(telegram_handler)
        else:
            self.logger.error("No Telegram credentials provided. Telegram logging will not work.")
    
    def get_logger(self):
        return self.logger
    
class TelegramBotHandler(logging.Handler):
    """
    This class is a custom logging handler that sends messages to a Telegram chat.
    """
    def __init__(self, token : str, chat_id : str, func : str):
        super().__init__()
        self.token = token
        self.chat_id = chat_id
        self.func = func
        self.display_new_betting_request()
    
    def send_msg(self, msg):
        if self.chat_id:
            apiURL = f"https://api.telegram.org/bot{self.token}/sendMessage"
            response = requests.post(
                apiURL, json={"chat_id": self.chat_id, "text": msg}
            )

    def format(self, record):
        emoji = self.get_emoji(record.levelname)
        if record.levelname == "INFO":
            return f"{emoji} {record.message}"
        return f"{emoji} {record.levelname} - {record.message}"

    def emit(self, record):
        log_entry = self.format(record)
        self.send_msg(log_entry)

    def display_new_betting_request(self):
        # Create a decorative header
        if self.func == "betting":
            header = (
                "✨✨✨✨✨✨✨✨✨✨\n"
                f"🕒 Betting Request Started 🕒\n"
                "✨✨✨✨✨✨✨✨✨✨\n"
            )
            # Send the header and messages
            self.send_msg(header)
            self.send_msg("🎉 New matches to bet on 🎉")
        if self.func == "retrieving":
            header = (
                "✨✨✨✨✨✨✨✨✨✨✨\n"
                f"🕒 Retrieving Request Started 🕒\n"
                "✨✨✨✨✨✨✨✨✨✨✨\n"
            )
            # Send the header and messages
            self.send_msg(header)
            self.send_msg("🎉 New matches to retrieve 🎉")

        self.send_msg("🔍 Let's see how it goes...")

    def get_emoji(self, levelname):
        emojis = {
            "DEBUG": "🐞",
            "INFO": "✅",
            "WARNING": "⚠️",
            "ERROR": "❌",
            "CRITICAL": "🔥"
        }
        return emojis.get(levelname, "ℹ️")




from datetime import datetime, timedelta
from typing import Any, Dict, List, Optional

from PS3838._PS3838Bet import Bet
from PS3838._PS3838Retrieve import Retrieve
from PS3838._telegram.telegram_bot import CustomLogger
from PS3838._utils.tools_check import check_config, check_list_matches
from PS3838._utils.tools_code import get_all_matches, place_bets

############################
#     Retrieving Class     #
############################

class PS3838RetrieveOdds():
    def __init__(
        self,
        config : Dict[str, str] = None,
        logger_active: Optional[bool] = True,
        logger_name: Optional[str] = "PS3838.log",
        *args,
        **kwargs
    ):
        self.config = config
        self.logger_active = logger_active
        self.logger_name = logger_name
        self.logger = None
        self.retrieve = None

        self._initiate()
        '''
        This class is used to retrieve the odds from the PS3838 API. It uses the Retrieve class to connect to the API and retrieve the odds for each match. It also uses a CustomLogger to log the information.
        
        Parameters:
            config (Dict[str, str]): The config to connect to the PS3838 API.
                Example: {"username": "my_username", "password": "my_password"}
            logger_active (bool): A boolean to activate the logger. Default is True.
            logger_name (str): The name of the log file. Default is "PS3838.log".
        '''

    def _initiate(self):
        # Create the logger. 
        if self.logger_active:
            custom_logger = CustomLogger(name="PS3838", log_file=self.logger_name, func="retrieving", config=self.config)
            self.logger = custom_logger.get_logger()

        # Check if the parameters are provided and valid (Raise a ConfigError if not)
        check_config(self.config)

        # Create the Retrieve object
        self.retrieve = Retrieve(config=self.config)

    ###############################
    #         Retrieving          #
    ###############################

    def retrieving(
        self,
        until_date : Optional[datetime] = None, 
        leagues : Optional[List[int]] = None
    ) -> List[List[Dict[str, Any]]]:
        """
        This function retrieves the odds for all the matches until a given date and for a given list of leagues. It uses the Retrieve class to connect to the PS3838 API and retrieve the odds for each match.

        Parameters:
            until_date (datetime): The date until which we want to retrieve the matches. Default is None, it retrieves everything in the specified leagues.
            leagues (List[int]): A list of leagues for which we want to retrieve the matches. Default is None, it retrieves every league until the date specified.

        Returns:
            List[List[Dict[str, Any]]]: A list of matches with their corresponding odds.
                Example: [{'id': 1595460299, 'starts': '2024-08-23T18:45:00Z', 'home': 'Paris Saint-Germain', 'away': 'Montpellier HSC', 'league': 2036, 'line_id': 2650184231, 'odds': {'home': 1.309, 'draw': 6.14, 'away': 8.47}}, ...]
        """

        try :
            matches = get_all_matches(self.retrieve, until_date, leagues, self.logger)
        except Exception as e:
            if self.logger:
                self.logger.error(f"Error while retrieving matches: {e}")

        return matches
    





############################
#      Betting Class       #
############################ 

class PS3838AutomatedBets():
    def __init__(
        self,
        config : Dict[str, str] = None,
        list_matches : List[Dict[str, Any]] = None,
        logger_active: Optional[bool] = True,
        to_bet: Optional[bool] = False,
        *args,
        **kwargs
    ):
        """
        This class is used to place bets on the PS3838 API. It uses the Bet classes to connect to the API and place the bets. It also uses a CustomLogger to log the information.
        
        Parameters:
            config (Dict[str, str]): The config to connect to the PS3838 API. 
                Example: {"username": "my_username", "password": "my_password"}
            list_matches (List[Dict[str, Any]]): A list of matches to bet on.
                Example: 
                    [{'id': 1596465583, 'starts': '2024-09-04T17:45:00Z', 'home': 'Resende FC', 'away': 'Artsul', 'league': 216059, 'line_id': 2673459336, 'odds': {'Team1': 2.05, 'Team2': 3.1, 'Draw': 3.25}, 'result': 1, 'amount': 5},
                    {'id': 1596465586, 'starts': '2024-09-04T17:45:00Z', 'home': 'Fluminense', 'away': 'AF Perolas Negras', 'league': 216059, 'line_id': 2673462355, 'odds': {'Team1': 1.159, 'Team2': 11.83, 'Draw': 5.83}, 'result': 0, 'amount': 5}, ...].
            logger_active (bool): A boolean to activate the logger. Default is True.
            to_bet (bool): A boolean to know if we really want to bets. Default is False, useful for testing the code.
        """

        self.config = config
        self.list_matches = list_matches
        self.logger_active = logger_active
        self.to_bet = to_bet
        self.bet = None
        self.logger = None

        self._initiate()

    def _initiate(self):
        # Create the logger. 
        if self.logger_active:
            custom_logger = CustomLogger(name="PS3838", log_file="PS3838.log", func="betting", config=self.config)
            self.logger = custom_logger.get_logger()

        # Check if the parameters are provided and valid (Raise a ConfigError if not), and check if the list of matches is valid (Raise a ParameterError if not)
        check_config(self.config)
        check_list_matches(self.list_matches)

        # Create the Bet and Retrieve objects
        self.bet = Bet(config=self.config)


    #############################
    #         Betting           #
    #############################

    def betting(self) -> List[Dict[str, Any]] :
        """
        This function places bets on the PS3838 API for a given list of matches, if the bets are not already placed and if the maintenance is not in progress.
 
        Returns:
            List[List[Dict[str, Any]]] | None: A list of matches with their corresponding odds if the bets were placed, None otherwise.
                Example: [{'id': 1595460299, 'starts': '2024-08-23T18:45:00Z', 'home': 'Paris Saint-Germain', 'away': 'Montpellier HSC', 'league': 2036, 'line_id': 2650184231, 'odds': {'home': 1.309, 'draw': 6.14, 'away': 8.47}}, ...]
        """
        matches_successful = None
        
        # Check if there is maintenance
        maintenance = self.bet.check_maintenance()

        if maintenance["status"] == 'ALL_BETTING_ENABLED':
            try:

                if self.to_bet:
                    # Place the bets
                    matches_successful = place_bets(self.list_matches, self.bet, logger=self.logger)
                else:
                    if self.logger:
                        self.logger.info("Bets not placed, to_bet is set to False")

            except Exception as e:
                if self.logger:
                    self.logger.error(f"Error placing bets: {e}")
        else:
            if self.logger:
                self.logger.info("Maintenance in progress, no bets placed. Try another time")

        return matches_successful 


import time
from concurrent.futures import ThreadPoolExecutor, as_completed
from datetime import datetime, timedelta
from typing import Any, Dict, List, Optional, Tuple

from PS3838._PS3838Bet import Bet
from PS3838._PS3838Retrieve import Retrieve
from PS3838._telegram.telegram_bot import CustomLogger
from PS3838._utils.constants import ID_SOCCER

######################################
#        Retrieving Functions        #
######################################

def get_all_matches(
        api_retrieve: Retrieve, 
        until_date: Optional[datetime] = None, 
        leagues_filter: Optional[List[int]] = None,
        logger: CustomLogger = None, 
    ) -> List[Dict[str, Any]]:
    """
    This function retrieves all the matches for the provided leagues and until the date specified. It also retrieves the odds for each match by parallelizing the process.

    Parameters:
        api_retrieve : Retrieve : API object to retrieve the odds
        until_date : Optional[datetime] : Date until which the matches are retrieved
        leagues_filter : Optional[List[int]] : Leagues to retrieve the matches from
        logger : CustomLogger : Logger to log the results

    Returns:
        match_odds : List[Dict[str, Any]] : List of matches with their corresponding odds
    """

    # Retrieve fixtures in the provided leagues or all of them if not provided
    all_fixtures = api_retrieve.get_fixtures_v3(sport_id=ID_SOCCER, league_ids=leagues_filter) if leagues_filter else api_retrieve.get_fixtures_v3(sport_id=ID_SOCCER)

    # Function to process each match (that we will parallelize)
    def process_match(league_id, match):
        try:
            # Check if match is open and (if specified) before the until_date
            if match["status"] == "O" and (not until_date or datetime.fromisoformat(match["starts"].replace('Z','')) <= until_date):
                match_info = {
                    "id": match["id"],
                    "starts": match["starts"],
                    "home": match["home"],
                    "away": match["away"],
                    "league": league_id,
                    "line_id": None
                }

                # Retrieve odds in the same step
                team1_odds, line_id = get_team_odds(api_retrieve, match_info, "Team1")
                team2_odds, _ = get_team_odds(api_retrieve, match_info, "Team2")
                draw_odds, _ = get_team_odds(api_retrieve, match_info, "Draw")

                return {**match_info, "line_id": line_id, "odds": {"home": team1_odds, "away": team2_odds, "draw": draw_odds}}
            
        except Exception as e:
            if logger:
                logger.error(f"Error processing match {match['id']}: {e}")

        return None

    # List to store the matches with odds
    match_odds = []  

    # Use a ThreadPoolExecutor to parallelize match processing
    with ThreadPoolExecutor(max_workers=10) as executor: 
        future_to_match = {
            executor.submit(process_match, league["id"], match): match
            for league in all_fixtures["league"]
            for match in league["events"]
        }

        # Collect results as they complete
        for future in as_completed(future_to_match):
            result = future.result()
            if result:
                match_odds.append(result)

    if logger:
        logger.info(f"Retrieved odds for {len(match_odds)} match(es)" if len(match_odds) > 0 else "No matches found")

    return match_odds
             

def get_team_odds(
    api_retrieve: Retrieve, 
    match: Dict[str, Any], 
    team: str
) -> Dict[str, Any]:
    """
    This function retrieves the odds for a specific team in a match.

    Parameters:
        api_retrieve : Retrieve : API object to retrieve the odds
        match : Dict[str, Any] : Match to retrieve the odds for
        team : str : Team to retrieve the odds for

    Returns:
        odds : Dict[str, Any] : Odds for the team
        line_id : int : Line ID of the bet
    """
    odds_response = api_retrieve.get_straight_line_v2(
        league_id=match["league"],
        handicap=0,
        odds_format="Decimal",
        sport_id=ID_SOCCER,
        event_id=match["id"],
        period_number=match.get("period_number", 0),
        bet_type="MONEYLINE",
        team=team
    )

    return odds_response.get("price", 0), odds_response.get("lineId", 0)



######################################
#          Betting Functions         #
######################################

def place_bets(
    list_matches : List[Tuple[Dict[str, Any], Dict[str, Any]]], 
    bet : Bet, 
    logger : CustomLogger = None,
) -> List[Dict[str, Any]]:
    """
    This function places the bets for the matches found with the odds provided. It verifies if the odds are not too low and if the bet is not already placed. Then in function of the parameter result, it places the bet on the home team, the away team or the draw.
    
    Parameters:
        list_matches : List[Tuple[Dict[str, Any], Dict[str, Any]]] : List of matches with their corresponding odds
            Example: [{'id': 1595460299, 'starts': '2024-08-23T18:45:00Z', 'home': 'Paris Saint-Germain', 'away': 'Montpellier HSC', 'league': 2036, 'line_id': 2650184231, 'odds': {'home': 1.309, 'draw': 6.14, 'away': 8.47}, 'result': 1, 'amount': 5}}, ...]
        bet : Bet : API object to place the bets
        logger : CustomLogger : Logger to log the results

    Returns:
        matches_successfull : List[Dict[str, Any]] : List of matches with their corresponding status
    """

    # List to store the matches with their status
    matches_successfull = []
    for match in list_matches:
        try : 
            # Check if the bet is already placed
            if is_already_bet(bet, match):
                if logger:
                    logger.warning(f"Bet already placed for match {match['id']} : {match['home']} - {match['away']}, skipping bet")
                matches_successfull.append({**match, "status": "ALREADY_BET"})
                continue
            # Place the bet depending on the "result" parameter
            if match["result"] == 1:
                bet.place_bet(stake=match["amount"], line_id=match["line_id"], sport_id=ID_SOCCER, event_id=match["id"], period_number=0, team="TEAM1")
            elif match["result"] == 2:
                bet.place_bet(stake=match["amount"], line_id=match["line_id"], sport_id=ID_SOCCER, event_id=match["id"], period_number=0, team="TEAM2")
            elif match["result"] == 0:
                bet.place_bet(stake=match["amount"], line_id=match["line_id"], sport_id=ID_SOCCER, event_id=match["id"], period_number=0, team="Draw")

            # Wait for 2 seconds to avoid rate limiting, and check if the bet is now well placed
            time.sleep(2)
            if is_already_bet(bet, match):
                if logger:
                    logger.info(f"Bet placed successfully for match : {match['home']} - {match['away']}")
                matches_successfull.append({**match, "status": "SUCCESS"})
                continue
            else:
                matches_successfull.append({**match, "status": "WTF"})

        except Exception as e:
            if logger:
                logger.error(f"Error placing bet for match {match['id']} : {match['home']} - {match['away']}: {e}")
            matches_successfull.append({**match, "status": "ERROR"})

    return matches_successfull

    
def change_date(date : str, days : int, hours : int) -> str:
    """
    This function changes the date by adding days and hours to it. It is useful to search for bets that were placed from 25 days in the past to 1 day in the future (as it is done in the function is_already_bet).
    """
    date = datetime.fromisoformat(date.replace('Z',''))
    date = date + timedelta(days=days, hours=hours)
    return date.isoformat() + 'Z'


def is_already_bet(api_bet : Bet, match : Dict[str, Any]) -> bool:
    """
    This function verifies if the bet is already placed for the match provided. It retrieves the bets placed in the past and checks if the bet is here. Also depending on the status of the bet, it returns True or False.
    """
    # Retrieve the bets placed in the past 25 days
    bets = api_bet.get_bets_v3(betlist='ALL', from_date=change_date(date=match["starts"], days=-25, hours=0), to_date=change_date(date=match["starts"], days=1, hours=0))

    # Check if the bet is already placed
    if "straightBets" in bets.keys():
        straights = bets["straightBets"]
        for bet in straights:
            if (bet["eventId"] == match["id"]) and (bet['betStatus'] == 'LOSE' or bet['betStatus'] == 'ACCEPTED' or bet['betStatus'] == 'CANCELLED' or bet['betStatus'] == 'WON' or bet['betStatus'] == 'REFUNDED' or bet['betStatus'] == 'PENDING_ACCEPTANCE'):
                return True
    
    return False




from datetime import datetime
from typing import Any, Dict, List, Optional

import Levenshtein

from PS3838._PS3838Retrieve import Retrieve
from PS3838._telegram.telegram_bot import CustomLogger
from PS3838._utils.constants import ID_SOCCER
from PS3838._utils.errors import RetrieveMatchError
from PS3838._utils.leagues_correspondency import CORRESPONDENCY_DICT_LEAGUES

    ##############################################
    #     Retrieve Odds For Specific Matches     #
    ##############################################

    # UNUSED FOR THE MOMENT

    # def retrieving(self, list_matches : List[Dict[str, Any]] = None) -> List[List[Dict[str, Any]]]:
    #     """
    #     This function retrieves the odds for a given list of matches. Several functions are used to find each match and their corresponding odds.
            
    #     Returns:
    #         List[List[Dict[str, Any]]]: A list of matches with their corresponding odds.
    #             Example: [({'id': 1595460299, 'starts': '2024-08-23T18:45:00Z', 'home': 'Paris Saint-Germain', 'away': 'Montpellier HSC', 'rotNum': '3121', 'liveStatus': 2, 'status': 'O', 'parlayRestriction': 2, 'altTeaser': False, 'resultingUnit': 'Regular', 'betAcceptanceType': 0, 'version': 545200449, 'league': 2036, 'result': 1, 'amount': 5, 'odd_min': 1.05, 'line_id': 2650184231}, {'team1_odds': 1.309, 'draw_odds': 6.14, 'team2_odds': 8.47}), ...]
    #     """

    #     # Retrieve the matches
    #     matches = retrieve_specific_matches(list_matches, self.retrieve, self.logger)

    #     # Retrieve the odds for each match
    #     match_odds = []
    #     for match in matches:
    #         try:
    #             team1_odds, match["line_id"] = get_team_odds(self.retrieve, match, "Team1")
    #             team2_odds, _ = get_team_odds(self.retrieve, match, "Team2")
    #             draw_odds, _ = get_team_odds(self.retrieve, match, "Draw")
                
    #             # Append the match and odds to the list "match_odds"
    #             match_odds.append((match, 
    #                 {
    #                     "team1_odds": team1_odds,
    #                     "draw_odds": draw_odds,
    #                     "team2_odds": team2_odds,
    #                 }
    #             ))
                
    #         except Exception as e:
    #             if self.logger:
    #                 self.logger.error(f"Error retrieving odds for match {match['event_id']}: {e}")
        
    #     if self.logger:
    #         self.logger.info(f"Retrieved odds for {len(match_odds)} match(es)" if len(match_odds) > 0 else "No matches found")
    #     return match_odds


def retrieve_specific_matches(
    list_matches : List[Dict[str, Any]],
    api_retrieve : Retrieve,
    logger : Optional[CustomLogger] = None,
    *args, 
    **kwargs
) -> List[Dict[str, Any]]:
    """
    This function retrieves the matches from the list of matches provided. It retrieves the matches one by one and appends them to a list of matches found.
    
    Parameters:
        list_matches : List[Dict[str, Any]] : List of matches to retrieve.
        api_retrieve : Retrieve : API object to retrieve the matches
        logger : Optional[CustomLogger] : Logger to log the results
 
    Returns:
        matches_found : List[Dict[str, Any]] : List of matches found
    """

    # Retrieve the matches
    matches_found = []
    for match in list_matches:
        
        try : 
            # Try to find the match that corresponds to the parameters
            match_found= retrieve_match(api_retrieve, league=match["league"], name_team1=match["team1"], name_team2=match["team2"], date=match["date"] if "date" in match.keys() else None, result=match["result"] if "result" in match.keys() else None, amount=match["amount"] if "amount" in match.keys() else None, odd_min=match["odd_min"] if "odd_min" in match.keys() else None, logger=logger)


            # Append the match to the list of matches found
            matches_found.append(match_found)
            if logger:
                logger.info(f"Match found for {match_found['home']} - {match_found['away']}")

        except Exception as e:
            if logger:
                logger.error(f"Error retrieving match for {match['team1']} - {match['team2']} : {e}")
            continue

    return matches_found

def retrieve_match(
    api_retrieve : Retrieve,
    league : int, 
    name_team1 : str, 
    name_team2 : str, 
    date : Optional[datetime] = None, 
    result : Optional[int] = None,
    amount : Optional[int] = None,
    odd_min : Optional[float] = None,
    logger : Optional[CustomLogger] = None,
    *args, 
    **kwargs
) -> Dict[str, Any]:
    """
    This function retrieves a match from the API based on the parameters provided. It retrieves the leagues, the matches in the league, and then finds the match that corresponds to the parameters.
    
    Parameters:
        api_retrieve : Retrieve : API object to retrieve the matches
        league : int : ID of the league
        name_team1 : str : Name of the first team
        name_team2 : str : Name of the second team
        date : Optional[datetime] : Date of the match
        result : Optional[int] : Result of the match (1 = team1 win, 2 = team2 win, 0 = draw)
        amount : Optional[int] : Amount to bet on the match
        odd_min : Optional[float] : Minimum odds to bet on the match
        logger : Optional[CustomLogger] : Logger to log the results
        
    Returns:
        match : Dict[str, Any] : Match found
    """

    # Find the ID of the league
    league_id = find_league_id_PS3838(league, logger)

    # Find the matches in the league
    fixtures = api_retrieve.get_fixtures_v3(sport_id=ID_SOCCER, league_ids=[league_id])

    # Keep only the elements where 'statut' = 'O' which means the match is open
    fixtures = [match for match in fixtures["league"][0]["events"] if match["status"] == "O"]

    print(fixtures)

    # Find the match
    match = find_match(fixtures, name_team1, name_team2, date, logger)

    # Verify if the parameter result is correct
    result = verify_result(match, name_team1, name_team2, result, logger)

    # Add the league ID to the match, the result, the amount and the minimum odds that are fundamentals parameters if we want to bet afterwards
    match["league"] = league_id
    match["result"] = result
    match["amount"] = amount
    match["odd_min"] = odd_min

    return match

def normalize_name(name : str) -> str:
    """
    This function normalizes a name by removing all special characters and converting it to lowercase.
    """
    return ''.join(e for e in name.lower() if e.isalnum())


def find_best_match(name: str, candidates: list[str], cutoff : float = 0.5) -> str:
    """
    This function finds the best match for a name in a list of candidates. It uses the Levenshtein distance to calculate the similarity between the name and the candidates. It is useful when the name of sport_api is not exactly the same as the name on PS3838. The cutoff parameter is used to filter the candidates that are too different from the name.
    """
    normalized_name = normalize_name(name)
    normalized_candidates = [normalize_name(candidate) for candidate in candidates]

    best_match = None
    highest_similarity = -1

    for candidate in normalized_candidates:
        similarity = Levenshtein.ratio(normalized_name, candidate)
        if similarity > highest_similarity and similarity >= cutoff:
            highest_similarity = similarity
            best_match = candidate

    if best_match:
        best_match_index = normalized_candidates.index(best_match)
        return candidates[best_match_index]

    return None

def find_match(
    fixtures : list[Dict[str, Any]], 
    name_team1 : str, 
    name_team2 : str, 
    date : datetime = None, 
    logger : CustomLogger = None
) -> Dict[str, Any]:
    """
    This function finds the match that corresponds to the team names and the date provided. It first tries to have the perfect match between team1 and team2 and the date, and then tries a lighter version without the date.
    
    Parameters:
        fixtures : list[Dict[str, Any]] : List of available matches returned by PS3838.
        name_team1 : str : Name of the first team
        name_team2 : str : Name of the second team
        date : datetime : Date of the match
        logger : CustomLogger : Logger to log the results
        
    Returns:
        match : Dict[str, Any] : Match found
    """
    # Find the candidates for the team names
    candidates_home = []
    candidates_away = []
    for match in fixtures:
        candidates_home.append(match["home"])
        candidates_away.append(match["away"])

    match = None
    # Find the match
    for fixture in fixtures:
        if (find_best_match(name_team1, candidates_home) == fixture["home"] and find_best_match(name_team2, candidates_away) == fixture["away"]) or (find_best_match(name_team1, candidates_away) == fixture["away"] and find_best_match(name_team2, candidates_home) == fixture["home"]) and datetime.fromisoformat(fixture["starts"].replace('Z','')) == date:
            match = fixture
            break
        elif (find_best_match(name_team1, candidates_home) == fixture["home"] and find_best_match(name_team2, candidates_away) == fixture["away"]) or (find_best_match(name_team1, candidates_away) == fixture["away"] and find_best_match(name_team2, candidates_home) == fixture["home"]):
            match = fixture
            break
        
    if not match:
        if logger:
            logger.error(f"Match not found for {name_team1} - {name_team2}")
        raise RetrieveMatchError("Match not found")
    
    return match


def verify_result(
    match : Dict[str, Any],
    name_team1 : str, 
    name_team2 : str, 
    result : int, 
    logger : CustomLogger = None
) -> int:
    """
    This function verifies if the result corresponds to the home team or the away team, and changes it if necessary.
    For example if we entered Marseille - Lens with result 1 and the actual match is Lens - Marseille, the function will return 2.
    
    Parameters:
        match : Dict[str, Any] : Match found
        name_team1 : str : Name of the first team
        name_team2 : str : Name of the second team
        result : int : Result of the match (1 = team1 win, 2 = team2 win, 0 = draw)
        logger : CustomLogger : Logger to log the results
        
    Returns:
        int : 1 if the result corresponds to the home team, 2 if the result corresponds to the away team, 0 if the result corresponds to a draw
    """
    if result is not None:
        if result == 1:
            if (find_best_match(match["home"], [name_team1, name_team2]) == name_team1) and (find_best_match(match["away"], [name_team1, name_team2]) == name_team2):
                return 1
            elif (find_best_match(match["away"], [name_team1, name_team2]) == name_team1) and (find_best_match(match["home"], [name_team1, name_team2]) == name_team2):
                return 2
            elif (find_best_match(match["home"], [name_team1, name_team2]) == name_team1) or (find_best_match(match["away"], [name_team1, name_team2]) == name_team2):
                return 1
            elif (find_best_match(match["away"], [name_team1, name_team2]) == name_team1) or (find_best_match(match["home"], [name_team1, name_team2]) == name_team2):
                return 2
            else:
                if logger:
                    logger.error(f"Error while verifying if the result 1 corresponds to the home team")
                raise RetrieveMatchError("Error while verifying if the result 1 corresponds to the home team")
        elif result == 2:
            if (find_best_match(match["home"], [name_team1, name_team2]) == name_team1) and (find_best_match(match["away"], [name_team1, name_team2]) == name_team2):
                return 2
            elif (find_best_match(match["away"], [name_team1, name_team2]) == name_team1) and (find_best_match(match["home"], [name_team1, name_team2]) == name_team2):
                return 1
            elif (find_best_match(match["home"], [name_team1, name_team2]) == name_team1) or (find_best_match(match["away"], [name_team1, name_team2]) == name_team2):
                return 2
            elif (find_best_match(match["away"], [name_team1, name_team2]) == name_team1) or (find_best_match(match["home"], [name_team1, name_team2]) == name_team2):
                return 1
            else:
                if logger:
                    logger.error(f"Error while verifying if the result 2 corresponds to the home team")
                raise RetrieveMatchError("Error while verifying if the result 2 corresponds to the home team")
        else:
            return 0
    else:
        return None
    
def find_league_id_PS3838(league_id_sport_api : int, logger : CustomLogger = None) -> int:
    """
    This function finds the league ID of PS3838 from the league ID of the sport API. It is useful to find the correct league to place the bets.
    
    Parameters:
        league_id_sport_api : int : League ID of the sport API
        logger : CustomLogger : Logger to log the results
        
    Returns:
        int : League ID of PS3838
    """
    for _, associations in CORRESPONDENCY_DICT_LEAGUES.items():
        for association_sport_api, association_ps3838 in associations.items():
            if association_sport_api[1] == league_id_sport_api:
                return association_ps3838[1]
    if logger:
        logger.error(f"League correspondency not found")
    raise RetrieveMatchError("League correspondency not found")